/* vim:expandtab:ts=2 sw=2:
*/
/*  Grafx2 - The Ultimate 256-color bitmap paint program

	Copyright owned by various GrafX2 authors, see COPYRIGHT.txt for details.

    Grafx2 is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; version 2
    of the License.

    Grafx2 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Grafx2; if not, see <http://www.gnu.org/licenses/>
*/
/* Sourcecode for this C64 binary is available in misc/c64viewer */
static unsigned char picview_prg[] = {
  0x01, 0x08, 0x0b, 0x08, 0x77, 0x07, 0x9e, 0x32, 0x30, 0x36, 0x31, 0x00,
  0x00, 0x00, 0xa9, 0x48, 0x85, 0xfb, 0xa9, 0x09, 0x85, 0xfc, 0xa2, 0x10,
  0xbd, 0x1f, 0xd0, 0x9d, 0xef, 0xc3, 0xca, 0xd0, 0xf7, 0xbd, 0x00, 0xd8,
  0x9d, 0x00, 0xc4, 0xbd, 0x00, 0xd9, 0x9d, 0x00, 0xc5, 0xbd, 0x00, 0xda,
  0x9d, 0x00, 0xc6, 0xbd, 0x00, 0xdb, 0x9d, 0x00, 0xc7, 0xe8, 0xd0, 0xe5,
  0x20, 0x21, 0x09, 0x48, 0x20, 0x21, 0x09, 0x48, 0x29, 0x0f, 0xaa, 0xbd,
  0x45, 0x09, 0x8d, 0x68, 0x08, 0x68, 0x29, 0xf0, 0xf0, 0x1c, 0x4a, 0x4a,
  0xaa, 0xbd, 0x31, 0x09, 0x85, 0xfd, 0xbd, 0x32, 0x09, 0x85, 0xfe, 0xbd,
  0x33, 0x09, 0x85, 0x02, 0xbd, 0x34, 0x09, 0xaa, 0x20, 0xfb, 0x08, 0x4c,
  0x3f, 0x08, 0x68, 0xaa, 0x09, 0x1b, 0x8d, 0x11, 0xd0, 0x8a, 0x29, 0x10,
  0x09, 0xc8, 0x8d, 0x16, 0xd0, 0xad, 0x00, 0xdd, 0x48, 0x29, 0xfc, 0x8d,
  0x00, 0xdd, 0xad, 0x18, 0xd0, 0x48, 0xa9, 0x08, 0x8d, 0x18, 0xd0, 0x20,
  0xe4, 0xff, 0xf0, 0xfb, 0xa9, 0x1b, 0x8d, 0x11, 0xd0, 0xa9, 0xc8, 0x8d,
  0x16, 0xd0, 0x68, 0x8d, 0x18, 0xd0, 0x68, 0x8d, 0x00, 0xdd, 0xa2, 0x10,
  0xbd, 0xef, 0xc3, 0x9d, 0x1f, 0xd0, 0xca, 0xd0, 0xf7, 0xbd, 0x00, 0xc4,
  0x9d, 0x00, 0xd8, 0xbd, 0x00, 0xc5, 0x9d, 0x00, 0xd9, 0xbd, 0x00, 0xc6,
  0x9d, 0x00, 0xda, 0xbd, 0x00, 0xc7, 0x9d, 0x00, 0xdb, 0xe8, 0xd0, 0xe5,
  0x60, 0x20, 0x21, 0x09, 0x20, 0x2c, 0x09, 0xc6, 0x02, 0xd0, 0x04, 0xca,
  0x10, 0x01, 0x60, 0x18, 0x69, 0x10, 0xb0, 0xed, 0x90, 0xee, 0x20, 0x21,
  0x09, 0xaa, 0x30, 0x06, 0xe8, 0x20, 0x0d, 0x09, 0xf0, 0xf4, 0xca, 0x10,
  0x30, 0x20, 0x23, 0x09, 0x20, 0x2c, 0x09, 0xe8, 0xd0, 0xfa, 0xf0, 0xe6,
  0xb1, 0xfb, 0x91, 0xfd, 0xc8, 0xd0, 0xf9, 0xe6, 0xfc, 0xe6, 0xfe, 0xca,
  0xd0, 0xf2, 0xa6, 0x02, 0xf0, 0x13, 0xb1, 0xfb, 0x91, 0xfd, 0xe6, 0xfb,
  0xd0, 0x02, 0xe6, 0xfc, 0xe6, 0xfd, 0xd0, 0x02, 0xe6, 0xfe, 0xca, 0xd0,
  0xed, 0x60, 0xa0, 0x00, 0xb1, 0xfb, 0xe6, 0xfb, 0xd0, 0x02, 0xe6, 0xfc,
  0x60, 0x91, 0xfd, 0xe6, 0xfd, 0xd0, 0x02, 0xe6, 0xfe, 0x60, 0x00, 0xe0,
  0x40, 0x1f, 0x00, 0xc0, 0xe8, 0x03, 0x00, 0xd8, 0xe8, 0x03, 0x20, 0xd0,
  0x0f, 0x00, 0xfb, 0xe1, 0xcc
};
